import abc

from rxconf import rxconf


class MetaObserver(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def get_actual_config(self) -> rxconf.RxConf:
        pass


class MetaAsyncObserver(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    async def get_actual_config(self) -> rxconf.RxConf:
        pass


class Observer(MetaObserver):

    def __init__(self, config: rxconf.RxConf):
        self._config = config

    def get_actual_config(self) -> rxconf.RxConf:
        # TODO: Fix stub
        return self._config


class AsyncObserver(MetaAsyncObserver):

    def __init__(self, config: rxconf.RxConf):
        self._config = config

    async def get_actual_config(self) -> rxconf.RxConf:
        # TODO: Fix stub
        return self._config



import functools
import typing as tp


class RxConfError(RuntimeError):
    def __init__(self, message: str):
        super().__init__(message)


class BrokenConfigSchemaError(RxConfError):
    def __init__(self, message: str):
        super().__init__(message)


class InvalidAttributeError(BrokenConfigSchemaError):
    def __init__(self, message: str):
        super().__init__(message)


class ConfigNotFoundError(RxConfError):
    def __init__(self, message: str):
        super().__init__(message)


class InvalidExtensionError(RxConfError):
    def __init__(self, message: str):
        super().__init__(message)


def handle_unknown_exception(func: tp.Callable[..., tp.Any]) -> tp.Callable[..., tp.Any]:
    @functools.wraps(func)
    def wrapper(*args: tp.Any, **kwargs: tp.Any) -> tp.Any:
        try:
            return func(*args, **kwargs)
        except RxConfError:
            raise
        except Exception as exc:
            raise RxConfError(f"An error occurred in {func.__name__}: {str(exc)}") from exc
    return wrapper



import typing as tp
from abc import ABCMeta, abstractmethod

from rxconf import exceptions, types


def _patch_other_value(func: tp.Callable[..., tp.Any]) -> tp.Callable[..., tp.Any]:
    def wrapper(self: "AttributeType", other: tp.Any) -> tp.Any:
        if isinstance(other, AttributeType):
            other = other.value
        return func(self, other)
    return wrapper


class AttributeType(metaclass=ABCMeta):
    _value: tp.Any

    def __init__(self, value: tp.Any) -> None:  # pragma: no cover
        self._value = value

    @property
    @abstractmethod
    def value(self) -> tp.Any:  # pragma: no cover
        raise NotImplementedError()

    @abstractmethod
    def __getattr__(self, item: str) -> tp.Any:  # pragma: no cover
        raise NotImplementedError()

    @exceptions.handle_unknown_exception
    def __int__(self) -> int:
        return int(self._value)  # type: ignore

    @exceptions.handle_unknown_exception
    def __float__(self) -> float:
        return float(self._value)  # type: ignore

    @exceptions.handle_unknown_exception
    def __str__(self) -> str:
        return str(self._value)  # type: ignore

    @exceptions.handle_unknown_exception
    def __bool__(self) -> bool:
        return bool(self._value)  # type: ignore

    @exceptions.handle_unknown_exception
    @_patch_other_value
    def __add__(self, other: tp.Any) -> tp.Any:
        return self._value + other  # type: ignore

    @exceptions.handle_unknown_exception
    @_patch_other_value
    def __mul__(self, other: tp.Any) -> tp.Any:
        return self._value * other  # type: ignore

    @exceptions.handle_unknown_exception
    @_patch_other_value
    def __sub__(self, other: tp.Any) -> tp.Any:
        return self._value - other  # type: ignore

    @exceptions.handle_unknown_exception
    @_patch_other_value
    def __truediv__(self, other: tp.Any) -> tp.Any:
        return self._value / other  # type: ignore

    @exceptions.handle_unknown_exception
    @_patch_other_value
    def __floordiv__(self, other: tp.Any) -> tp.Any:
        return self._value // other  # type: ignore

    @exceptions.handle_unknown_exception
    @_patch_other_value
    def __mod__(self, other: tp.Any) -> tp.Any:
        return self._value % other  # type: ignore

    @exceptions.handle_unknown_exception
    @_patch_other_value
    def __pow__(self, other: tp.Any) -> tp.Any:
        return self._value ** other  # type: ignore

    @exceptions.handle_unknown_exception
    @_patch_other_value
    def __lshift__(self, other: tp.Any) -> tp.Any:
        return self._value << other  # type: ignore

    @exceptions.handle_unknown_exception
    @_patch_other_value
    def __rshift__(self, other: tp.Any) -> tp.Any:
        return self._value >> other  # type: ignore

    @exceptions.handle_unknown_exception
    @_patch_other_value
    def __and__(self, other: tp.Any) -> tp.Any:
        return self._value & other  # type: ignore

    @exceptions.handle_unknown_exception
    @_patch_other_value
    def __xor__(self, other: tp.Any) -> tp.Any:
        return self._value ^ other  # type: ignore

    @exceptions.handle_unknown_exception
    @_patch_other_value
    def __or__(self, other: tp.Any) -> tp.Any:
        return self._value | other  # type: ignore

    @exceptions.handle_unknown_exception
    def __neg__(self) -> tp.Any:
        return -self._value  # type: ignore

    @exceptions.handle_unknown_exception
    def __pos__(self) -> tp.Any:
        return +self._value  # type: ignore

    @exceptions.handle_unknown_exception
    def __abs__(self) -> tp.Any:
        return abs(self._value)  # type: ignore

    @exceptions.handle_unknown_exception
    def __invert__(self) -> tp.Any:
        return ~self._value  # type: ignore

    @exceptions.handle_unknown_exception
    def __round__(self, n: int = 0) -> tp.Any:
        return round(self._value, n)  # type: ignore

    @exceptions.handle_unknown_exception
    def __hash__(self) -> int:
        return hash(self._value)  # type: ignore

    @exceptions.handle_unknown_exception
    def __len__(self) -> int:
        return len(self._value)  # type: ignore

    @exceptions.handle_unknown_exception
    def __iter__(self) -> tp.Iterator[tp.Any]:
        return iter(self._value)  # type: ignore

    @exceptions.handle_unknown_exception
    @_patch_other_value
    def __contains__(self, item: tp.Any) -> bool:
        return item in self._value  # type: ignore

    @exceptions.handle_unknown_exception
    @_patch_other_value
    def __getitem__(self, item: tp.Any) -> tp.Any:
        return self._value[item]  # type: ignore

    @exceptions.handle_unknown_exception
    @_patch_other_value
    def __eq__(self, other: tp.Any) -> bool:
        return self._value == other  # type: ignore

    @exceptions.handle_unknown_exception
    @_patch_other_value
    def __ne__(self, other: tp.Any) -> bool:
        return not self.__eq__(other)  # type: ignore

    @exceptions.handle_unknown_exception
    @_patch_other_value
    def __lt__(self, other: tp.Any) -> bool:
        return self._value < other  # type: ignore

    @exceptions.handle_unknown_exception
    @_patch_other_value
    def __le__(self, other: tp.Any) -> bool:
        return self._value <= other  # type: ignore

    @exceptions.handle_unknown_exception
    @_patch_other_value
    def __gt__(self, other: tp.Any) -> bool:
        return other < self._value  # type: ignore

    @exceptions.handle_unknown_exception
    @_patch_other_value
    def __ge__(self, other: tp.Any) -> bool:
        return other <= self._value  # type: ignore

    @exceptions.handle_unknown_exception
    def __repr__(self) -> str:
        return f"AttributeType({self._value})"


class MockAttribute(AttributeType):  # pragma: no cover
    _value: tp.Any

    def __init__(self, value: tp.Any = None) -> None:
        self._value = value

    @property
    def value(self) -> None:
        pass

    def __getattr__(self, item: str) -> None:
        pass


class YamlAttribute(AttributeType):
    _value: tp.Union[
        types.YAML_ATTRIBUTE_TYPE,
        "YamlAttribute",
        tp.List["YamlAttribute"],
        tp.Set["YamlAttribute"],
        tp.Dict[str, "YamlAttribute"],
    ]

    def __init__(
        self: "YamlAttribute",
        value: tp.Union[
            types.YAML_ATTRIBUTE_TYPE,
            "YamlAttribute",
            tp.List["YamlAttribute"],
            tp.Set["YamlAttribute"],
            tp.Dict[str, "YamlAttribute"],
        ]
    ) -> None:
        self._value = value

    @property
    def value(
        self: "YamlAttribute"
    ) -> tp.Union[
        types.YAML_ATTRIBUTE_TYPE,
        "YamlAttribute",
        tp.List["YamlAttribute"],
        tp.Set["YamlAttribute"],
        tp.Dict[str, "YamlAttribute"],
    ]:
        return self._value

    @exceptions.handle_unknown_exception
    def __getattr__(
        self: "YamlAttribute",
        item: str,
    ) -> tp.Union[
        types.YAML_ATTRIBUTE_TYPE,
        "YamlAttribute",
        tp.List["YamlAttribute"],
        tp.Set["YamlAttribute"],
        tp.Dict[str, "YamlAttribute"],
    ]:
        if isinstance(self._value, dict):
            try:
                return self._value[item.lower()]
            except KeyError as exc:
                raise KeyError(f"Key `{item}` doesn't exist...") from exc

        raise KeyError(f"Key `{item}` doesn't exist...")


class JsonAttribute(AttributeType):
    _value: tp.Union[
        types.JSON_ATTRIBUTE_TYPE,
        "JsonAttribute",
        tp.List["JsonAttribute"],
        tp.Dict[str, "JsonAttribute"],
    ]

    def __init__(
        self: "JsonAttribute",
        value: tp.Union[
            types.JSON_ATTRIBUTE_TYPE,
            "JsonAttribute",
            tp.List["JsonAttribute"],
            tp.Dict[str, "JsonAttribute"],
        ]
    ) -> None:
        self._value = value

    @property
    def value(
        self: "JsonAttribute"
    ) -> tp.Union[
        types.JSON_ATTRIBUTE_TYPE,
        "JsonAttribute",
        tp.List["JsonAttribute"],
        tp.Dict[str, "JsonAttribute"],
    ]:
        return self._value

    @exceptions.handle_unknown_exception
    def __getattr__(
        self: "JsonAttribute",
        item: str,
    ) -> tp.Union[
        types.JSON_ATTRIBUTE_TYPE,
        "JsonAttribute",
        tp.List["JsonAttribute"],
        tp.Dict[str, "JsonAttribute"],
    ]:
        if isinstance(self._value, dict):
            try:
                return self._value[item.lower()]
            except KeyError as exc:
                raise KeyError(f"Key `{item}` doesn't exist...") from exc

        raise KeyError(f"Key `{item}` doesn't exist...")


class TomlAttribute(AttributeType):
    _value: tp.Union[
        types.TOML_ATTRIBUTE_TYPE,
        "TomlAttribute",
        tp.List["TomlAttribute"],
        tp.Dict[str, "TomlAttribute"],
    ]

    def __init__(
        self: "TomlAttribute",
        value: tp.Union[
            types.TOML_ATTRIBUTE_TYPE,
            "TomlAttribute",
            tp.List["TomlAttribute"],
            tp.Dict[str, "TomlAttribute"],
        ]
    ) -> None:
        self._value = value

    @property
    def value(
        self: "TomlAttribute"
    ) -> tp.Union[
        types.TOML_ATTRIBUTE_TYPE,
        "TomlAttribute",
        tp.List["TomlAttribute"],
        tp.Dict[str, "TomlAttribute"],
    ]:
        return self._value

    @exceptions.handle_unknown_exception
    def __getattr__(
        self: "TomlAttribute",
        item: str,
    ) -> tp.Union[
        types.TOML_ATTRIBUTE_TYPE,
        "TomlAttribute",
        tp.List["TomlAttribute"],
        tp.Dict[str, "TomlAttribute"],
    ]:
        if isinstance(self._value, dict):
            try:
                return self._value[item.lower()]
            except KeyError as exc:
                raise KeyError(f"Key `{item}` doesn't exist...") from exc

        raise KeyError(f"Key `{item}` doesn't exist...")


class EnvAttribute(AttributeType):
    _value: tp.Union[
        types.ENV_ATTRIBUTE_TYPE,
        "EnvAttribute",
        tp.Dict[str, "EnvAttribute"],
    ]

    def __init__(
        self: "EnvAttribute",
        value: tp.Union[
            types.ENV_ATTRIBUTE_TYPE,
            "EnvAttribute",
            tp.Dict[str, "EnvAttribute"],
        ]
    ) -> None:
        self._value = value

    @property
    def value(
        self: "EnvAttribute"
    ) -> tp.Union[
        types.ENV_ATTRIBUTE_TYPE,
        "EnvAttribute",
        tp.Dict[str, "EnvAttribute"],
    ]:
        return self._value

    @exceptions.handle_unknown_exception
    def __getattr__(
        self: "EnvAttribute",
        item: str,
    ) -> tp.Union[
        types.ENV_ATTRIBUTE_TYPE,
        "EnvAttribute",
        tp.Dict[str, "EnvAttribute"],
    ]:
        if isinstance(self._value, dict):
            try:
                return self._value[item.lower()]
            except KeyError as exc:
                raise KeyError(f"Key `{item}` doesn't exist...") from exc

        raise KeyError(f"Key `{item}` doesn't exist...")


class IniAttribute(AttributeType):
    _value: tp.Union[
        types.INI_ATTRIBUTE_TYPE,
        tp.Dict[str, "IniAttribute"],
    ]

    def __init__(self: "IniAttribute", value: tp.Union[
        types.INI_ATTRIBUTE_TYPE,
        tp.Dict[str, "IniAttribute"]
    ]) -> None:
        self._value = value

    @property
    def value(
            self: "IniAttribute"
    ) -> tp.Union[
        types.INI_ATTRIBUTE_TYPE,
        tp.Dict[str, "IniAttribute"]
    ]:
        return self._value

    @exceptions.handle_unknown_exception
    def __getattr__(
            self: "IniAttribute",
            item: str,
    ) -> tp.Union[
        types.INI_ATTRIBUTE_TYPE,
        tp.Dict[str, "IniAttribute"],
        "IniAttribute"
    ]:
        if isinstance(self._value, dict):
            try:
                return self._value[item.lower()]
            except KeyError as exc:
                raise KeyError(f"Key `{item}` doesn't exist...") from exc

        raise KeyError(f"Key `{item}` doesn't exist...")



import abc
import pathlib
import typing as tp
from rxconf.config_resolver import ConfigResolver, FileConfigResolver
from rxconf.config_types import ConfigType, FileConfigType


class ConfigBuilder(metaclass=abc.ABCMeta):  # pragma: no cover

    def __init__(self, config_resolver: ConfigResolver) -> None:
        self._config_resolver = config_resolver

    @abc.abstractmethod
    def build(self, *args, **kwargs) -> ConfigType:
        pass


class FileConfigBuilder(ConfigBuilder):

    def __init__(self, config_resolver: FileConfigResolver) -> None:
        self._config_resolver: FileConfigResolver = config_resolver

    def build(
        self: "FileConfigBuilder",
        path: tp.Union[str, pathlib.PurePath],
        encoding: str,
    ) -> FileConfigType:
        return self._config_resolver.resolve(path=path).load_from_path(
            path=path,
            encoding=encoding,
        )

    async def build_async(
        self,
        path: tp.Union[str, pathlib.PurePath],
        encoding: str,
    ) -> FileConfigType:
        return await self._config_resolver.resolve(path=path).load_from_path_async(
            path=path,
            encoding=encoding,
        )



import os
import pathlib
import typing as tp
from abc import ABCMeta

from rxconf import attributes, config_types, exceptions


class ConfigResolver(metaclass=ABCMeta):
    pass


class FileConfigResolver(ConfigResolver):

    def __init__(self, config_types: tp.Iterable[tp.Type[config_types.FileConfigType]]) -> None:
        self._config_types = config_types

    def resolve(
        self,
        path: tp.Union[str, pathlib.PurePath],
    ) -> tp.Type[config_types.FileConfigType]:
        _, extension = os.path.splitext(path)
        for config_type in self._config_types:
            if extension in config_type(
                root_attribute=attributes.MockAttribute(),
                path=pathlib.Path(),
            ).allowed_extensions:
                return config_type

        # TODO: add here link how to patch the extensions.
        raise exceptions.InvalidExtensionError(
            f"Config file `{path}` has invalid extension: {extension}. "
            f"If you want to support this extension, "
            f"follow the tiny guideline: ..."
        )


DefaultFileConfigResolver: tp.Final[FileConfigResolver] = FileConfigResolver(
    config_types=config_types.BASE_FILE_CONFIG_TYPES,
)



import sys
import typing as tp
from datetime import date, datetime

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

YAML_ATTRIBUTE_TYPE: TypeAlias = tp.Union[
    tp.Union[bool, int, str, float, None],
    tp.List[tp.Union[bool, int, str, float, None]],
    tp.Set[tp.Union[bool, int, str, float, None]],
    tp.Union[date, datetime],
]

JSON_ATTRIBUTE_TYPE: TypeAlias = tp.Union[
    tp.Union[bool, int, str, float, None],
    tp.List[tp.Union[bool, int, str, float, None]],
]

TOML_ATTRIBUTE_TYPE: TypeAlias = tp.Union[
    tp.Union[bool, int, str, float],
    tp.List[tp.Union[bool, int, str, float]],
    tp.Union[date, datetime],
]

INI_ATTRIBUTE_TYPE: TypeAlias = tp.Union[
    bool, int, str, float, None,
]

ENV_ATTRIBUTE_TYPE: TypeAlias = tp.Union[
    bool, int, str, float, None,
]


def map_primitive(value: str) -> tp.Union[int, float, bool, None, str]:
    lower_value = value.lower()
    if lower_value == "none" or lower_value == "null":
        return None
    if lower_value == "true":
        return True
    if lower_value == "false":
        return False
    try:
        return int(value)
    except ValueError:
        pass
    try:
        return float(value)
    except ValueError:
        pass
    return value



from rxconf import types

from .attributes import (
    AttributeType,
    YamlAttribute,
    JsonAttribute,
    TomlAttribute,
    IniAttribute,
    EnvAttribute,
)
from .config_resolver import ConfigResolver
from .config_types import (
    YamlConfig,
    JsonConfig,
    TomlConfig,
    IniConfig,
    EnvConfig,
    DotenvConfig,
)
from .exceptions import (
    BrokenConfigSchemaError,
    ConfigNotFoundError,
    InvalidExtensionError,
)
from .rxconf import (
    RxConf,
    AsyncRxConf,
)

__all__ = [
    "RxConf",
    "AsyncRxConf",
    "AttributeType",
    "YamlAttribute",
    "JsonAttribute",
    "TomlAttribute",
    "IniAttribute",
    "EnvAttribute",
    "ConfigResolver",
    "YamlConfig",
    "JsonConfig",
    "TomlConfig",
    "IniConfig",
    "EnvConfig",
    "DotenvConfig",
    "BrokenConfigSchemaError",
    "ConfigNotFoundError",
    "InvalidExtensionError",
    "types",
    "attributes",
    "config_resolver",
    "config_types",
    "exceptions",
]



import datetime
import json
import os
import sys
import typing as tp
from abc import ABCMeta, abstractmethod
from pathlib import PurePath
import aiofiles
import yaml
from dotenv import load_dotenv

from rxconf import types


if sys.version_info >= (3, 11):
    import tomllib as toml
else:
    import toml

import configparser

import rxconf
from rxconf import attributes as attrs
from rxconf import exceptions


class ConfigType(metaclass=ABCMeta):  # pragma: no cover

    @abstractmethod
    def __getattr__(self, item: str) -> tp.Any:
        raise NotImplementedError()


class FileConfigType(ConfigType, metaclass=ABCMeta):  # pragma: no cover

    def __init__(
            self: "FileConfigType",
            root_attribute: rxconf.AttributeType,
            path: PurePath,
    ) -> None:
        pass

    @property
    @abstractmethod
    def allowed_extensions(self) -> tp.FrozenSet[str]:
        pass

    @classmethod
    @abstractmethod
    def load_from_path(
        cls,
        path: tp.Union[str, PurePath],
        encoding: str,
    ) -> "FileConfigType":
        pass

    @classmethod
    @abstractmethod
    async def load_from_path_async(
        cls,
        path: tp.Union[str, PurePath],
        encoding: str,
    ) -> "FileConfigType":
        pass

    def __repr__(self) -> str:
        return repr(self._root)


class YamlConfig(FileConfigType):
    _allowed_extensions: tp.Final[frozenset] = frozenset({".yaml", ".yml"})
    _root: tp.Final[rxconf.YamlAttribute]
    _path: tp.Final[PurePath]

    def __init__(
            self: "YamlConfig",
            root_attribute: rxconf.YamlAttribute,
            path: PurePath,
    ) -> None:
        self._root = root_attribute
        self._path = path

    @property
    def allowed_extensions(self) -> tp.FrozenSet[str]:
        return self._allowed_extensions

    @classmethod
    def _load_yaml_data(cls, content: str, path: tp.Union[str, PurePath]) -> tp.Dict:
        try:
            return yaml.safe_load(content)
        except yaml.YAMLError as exc:
            raise exceptions.BrokenConfigSchemaError(
                f"Error while parsing yaml config: {path}"
            ) from exc

    @classmethod
    @exceptions.handle_unknown_exception
    def load_from_path(
        cls,
        path: tp.Union[str, PurePath],
        encoding: str,
    ) -> "YamlConfig":
        if not os.path.isfile(str(path)):
            raise exceptions.ConfigNotFoundError(f"Config file not found: {path}")
        with open(str(path), encoding=encoding) as file:
            yaml_data = cls._load_yaml_data(file.read(), path)

        return cls(
            root_attribute=cls._process_data(yaml_data),
            path=path if isinstance(path, PurePath) else PurePath(path)
        )

    @classmethod
    @exceptions.handle_unknown_exception
    async def load_from_path_async(
        cls,
        path: tp.Union[str, PurePath],
        encoding: str,
    ) -> "YamlConfig":
        if not os.path.isfile(str(path)):
            raise exceptions.ConfigNotFoundError(f"Config file not found: {path}")
        async with aiofiles.open(str(path), mode="r", encoding=encoding) as file:
            content = await file.read()
            yaml_data = cls._load_yaml_data(content, path)

        return cls(
            root_attribute=cls._process_data(yaml_data),
            path=path if isinstance(path, PurePath) else PurePath(path)
        )

    @exceptions.handle_unknown_exception
    def __getattr__(self, item: str) -> tp.Any:
        return getattr(self._root, item.lower())

    @classmethod
    @exceptions.handle_unknown_exception
    def _process_data(cls, data: tp.Any) -> attrs.YamlAttribute:
        if isinstance(data, tp.Dict):
            return attrs.YamlAttribute(
                value={k.lower(): cls._process_data(v) for k, v in data.items()}
            )
        elif isinstance(data, tp.List):
            return attrs.YamlAttribute(
                value=[
                    cls._process_data(item) if not isinstance(item, dict) else attrs.YamlAttribute(
                        value={k.lower(): cls._process_data(v) for k, v in item.items()}
                    ) for item in data
                ]
            )
        elif isinstance(data, tp.Set):
            return attrs.YamlAttribute(  # pragma: no cover
                value={cls._process_data(item) for item in data}
            )
        elif isinstance(data, (bool, int, str, float, type(None), datetime.date, datetime.datetime)):
            return attrs.YamlAttribute(value=data)
        else:
            raise exceptions.BrokenConfigSchemaError(f"Unsupported data type: {type(data)}")  # pragma: no cover


class JsonConfig(FileConfigType):
    _allowed_extensions: tp.Final[frozenset] = frozenset({".json"})
    _root: tp.Final[rxconf.JsonAttribute]
    _path: tp.Final[PurePath]

    def __init__(
            self: "JsonConfig",
            root_attribute: rxconf.JsonAttribute,
            path: PurePath,
    ) -> None:
        self._root = root_attribute
        self._path = path

    @property
    def allowed_extensions(self) -> tp.FrozenSet[str]:
        return self._allowed_extensions

    @classmethod
    def _load_json_data(cls, content: str, path: tp.Union[str, PurePath]) -> tp.Dict:
        try:
            return json.loads(content)
        except json.JSONDecodeError as exc:
            raise exceptions.BrokenConfigSchemaError(
                f"Error while parsing json config: {path}"
            ) from exc

    @classmethod
    @exceptions.handle_unknown_exception
    def load_from_path(
        cls,
        path: tp.Union[str, PurePath],
        encoding: str,
    ) -> "JsonConfig":
        if not os.path.isfile(str(path)):
            raise exceptions.ConfigNotFoundError(f"Config file not found: {path}")
        with open(str(path), encoding=encoding) as file:
            content = file.read()
            json_data = cls._load_json_data(content, path)

        return cls(
            root_attribute=cls._process_data(json_data),
            path=path if isinstance(path, PurePath) else PurePath(path)
        )

    @classmethod
    @exceptions.handle_unknown_exception
    async def load_from_path_async(
        cls,
        path: tp.Union[str, PurePath],
        encoding: str,
    ) -> "JsonConfig":
        if not os.path.isfile(str(path)):
            raise exceptions.ConfigNotFoundError(f"Config file not found: {path}")
        async with aiofiles.open(str(path), mode="r", encoding=encoding) as file:
            content = await file.read()
            json_data = cls._load_json_data(content, path)

        return cls(
            root_attribute=cls._process_data(json_data),
            path=path if isinstance(path, PurePath) else PurePath(path)
        )

    @exceptions.handle_unknown_exception
    def __getattr__(self, item: str) -> tp.Any:
        return getattr(self._root, item.lower())

    @classmethod
    @exceptions.handle_unknown_exception
    def _process_data(cls, data: tp.Any) -> attrs.JsonAttribute:
        if isinstance(data, dict):
            return attrs.JsonAttribute(
                value={k.lower(): cls._process_data(v) for k, v in data.items()}
            )
        elif isinstance(data, list):
            return attrs.JsonAttribute(
                value=[
                    cls._process_data(item) if not isinstance(item, dict) else attrs.JsonAttribute(
                        value={k.lower(): cls._process_data(v) for k, v in item.items()}
                    ) for item in data
                ]
            )
        elif isinstance(data, (bool, int, str, float, type(None))):
            return attrs.JsonAttribute(value=data)
        else:
            raise exceptions.BrokenConfigSchemaError(f"Unsupported data type: {type(data)}")  # pragma: no cover


class TomlConfig(FileConfigType):
    _allowed_extensions: tp.Final[frozenset] = frozenset({".toml"})
    _root: tp.Final[rxconf.TomlAttribute]
    _path: tp.Final[PurePath]

    def __init__(
            self: "TomlConfig",
            root_attribute: rxconf.TomlAttribute,
            path: PurePath,
    ) -> None:
        self._root = root_attribute
        self._path = path

    @property
    def allowed_extensions(self) -> tp.FrozenSet[str]:
        return self._allowed_extensions

    @classmethod
    def _load_toml_data(cls, content: str, path: tp.Union[str, PurePath]) -> tp.Dict:
        toml_decode_exc = (
            toml.TOMLDecodeError  # type: ignore
            if sys.version_info >= (3, 11)
            else toml.TomlDecodeError  # type: ignore
        )
        try:
            if sys.version_info >= (3, 11):
                return toml.loads(content)
            else:
                return toml.loads(content)  # pragma: no cover
        except toml_decode_exc as exc:
            raise exceptions.BrokenConfigSchemaError(
                f"Error while parsing toml config: {path}"
            ) from exc

    @classmethod
    @exceptions.handle_unknown_exception
    def load_from_path(
        cls,
        path: tp.Union[str, PurePath],
        encoding: str,
    ) -> "TomlConfig":
        if not os.path.isfile(str(path)):
            raise exceptions.ConfigNotFoundError(f"Config file not found: {path}")
        with open(str(path), "r", encoding=encoding) as file:
            content = file.read()
            toml_data = cls._load_toml_data(content, path)

        return cls(
            root_attribute=cls._process_data(toml_data),
            path=path if isinstance(path, PurePath) else PurePath(path)
        )

    @classmethod
    @exceptions.handle_unknown_exception
    async def load_from_path_async(
        cls,
        path: tp.Union[str, PurePath],
        encoding: str,
    ) -> "TomlConfig":
        if not os.path.isfile(str(path)):
            raise exceptions.ConfigNotFoundError(f"Config file not found: {path}")
        async with aiofiles.open(str(path), "r", encoding=encoding) as file:
            content = await file.read()
            toml_data = cls._load_toml_data(content, path)

        return cls(
            root_attribute=cls._process_data(toml_data),
            path=path if isinstance(path, PurePath) else PurePath(path)
        )

    @exceptions.handle_unknown_exception
    def __getattr__(self, item: str) -> tp.Any:
        return getattr(self._root, item.lower())

    @classmethod
    @exceptions.handle_unknown_exception
    def _process_data(cls, data: tp.Any) -> attrs.TomlAttribute:
        if isinstance(data, dict):
            return attrs.TomlAttribute(
                value={k.lower(): cls._process_data(v) for k, v in data.items()}
            )
        elif isinstance(data, list):
            return attrs.TomlAttribute(
                value=[
                    cls._process_data(item) if not isinstance(item, dict) else attrs.TomlAttribute(
                        value={k.lower(): cls._process_data(v) for k, v in item.items()}
                    ) for item in data
                ]
            )
        elif isinstance(data, (bool, int, str, float, datetime.date, datetime.datetime)):
            return attrs.TomlAttribute(value=data)
        else:
            raise exceptions.BrokenConfigSchemaError(f"Unsupported data type: {type(data)}")  # pragma: no cover


class IniConfig(FileConfigType):
    _allowed_extensions: tp.Final[frozenset] = frozenset({".ini"})
    _root: tp.Final[rxconf.IniAttribute]
    _path: tp.Final[PurePath]

    def __init__(
            self: "IniConfig",
            root_attribute: rxconf.IniAttribute,
            path: PurePath,
    ) -> None:
        self._root = root_attribute
        self._path = path

    @property
    def allowed_extensions(self) -> tp.FrozenSet[str]:
        return self._allowed_extensions

    @classmethod
    def _load_ini_data(cls, content: str, path: tp.Union[str, PurePath]) -> tp.Dict:
        config = configparser.ConfigParser()
        try:
            config.read_string(content)
        except configparser.Error as exc:
            raise exceptions.BrokenConfigSchemaError(
                f"Error while parsing ini config: {path}"
            ) from exc

        ini_data: tp.Dict[tp.Any, tp.Any] = dict()
        for section in config.sections():
            keys = section.split('.')
            current_level = ini_data
            for key in keys[:-1]:
                if key not in current_level:
                    current_level[key] = dict()
                current_level = current_level[key]
            current_level[keys[-1]] = dict(config.items(section))
        return ini_data

    @classmethod
    @exceptions.handle_unknown_exception
    def load_from_path(
        cls,
        path: tp.Union[str, PurePath],
        encoding: str,
    ) -> "IniConfig":
        if not os.path.isfile(str(path)):
            raise exceptions.ConfigNotFoundError(f"Config file not found: {path}")
        with open(str(path), mode="r", encoding=encoding) as file:
            content = file.read()
            ini_data = cls._load_ini_data(content, path)

        return cls(
            root_attribute=cls._process_data(ini_data),
            path=path if isinstance(path, PurePath) else PurePath(path),
        )

    @classmethod
    @exceptions.handle_unknown_exception
    async def load_from_path_async(
        cls,
        path: tp.Union[str, PurePath],
        encoding: str,
    ) -> "IniConfig":
        if not os.path.isfile(str(path)):
            raise exceptions.ConfigNotFoundError(f"Config file not found: {path}")
        async with aiofiles.open(str(path), mode="r", encoding=encoding) as file:
            content = await file.read()
            ini_data = cls._load_ini_data(content, path)

        return cls(
            root_attribute=cls._process_data(ini_data),
            path=path if isinstance(path, PurePath) else PurePath(path),
        )

    @exceptions.handle_unknown_exception
    def __getattr__(self, item: str) -> tp.Any:
        return getattr(self._root, item.lower())

    @classmethod
    @exceptions.handle_unknown_exception
    def _process_data(cls, data: tp.Any) -> attrs.IniAttribute:
        if isinstance(data, dict):
            return attrs.IniAttribute(
                value={k.lower(): cls._process_data(v) for k, v in data.items()}
            )
        elif isinstance(data, str):
            return attrs.IniAttribute(value=types.map_primitive(data))
        else:
            raise exceptions.BrokenConfigSchemaError(f"Unsupported data type: {type(data)}")  # pragma: no cover


class EnvConfig(ConfigType):

    _root: tp.Final[rxconf.EnvAttribute]

    def __init__(self: "EnvConfig", root_attribute: rxconf.EnvAttribute) -> None:
        self._root = root_attribute

    def __repr__(self) -> str:
        return repr(self._root)

    @exceptions.handle_unknown_exception
    def __getattr__(self, item: str) -> tp.Any:
        return getattr(self._root, item.lower())

    @classmethod
    @exceptions.handle_unknown_exception
    def load_from_environment(
            cls,
            prefix: tp.Optional[str] = None,
            remove_prefix: tp.Optional[bool] = False,
    ) -> "EnvConfig":
        if prefix and remove_prefix:
            env_vars = {
                k.lower().removeprefix(prefix.lower()).lstrip('_'): v
                for k, v in os.environ.items()
                if k.lower().startswith(prefix.lower())
            }
        elif prefix:
            env_vars = {
                k.lower(): v
                for k, v in os.environ.items()
                if k.lower().startswith(prefix.lower())
            }
        else:
            env_vars = {k.lower(): v for k, v in os.environ.items()}
        root_attribute = cls._process_data(env_vars)
        return cls(root_attribute=root_attribute)

    @classmethod
    @exceptions.handle_unknown_exception
    def _process_data(cls, data: tp.Dict[str, str]) -> attrs.EnvAttribute:
        processed_data = {k.lower(): attrs.EnvAttribute(value=types.map_primitive(v)) for k, v in data.items()}
        return attrs.EnvAttribute(value=processed_data)


class DotenvConfig(FileConfigType, EnvConfig):

    _allowed_extensions: tp.Final[frozenset] = frozenset({".env"})
    _path: tp.Final[PurePath]

    def __init__(
        self: "DotenvConfig",
        root_attribute: rxconf.EnvAttribute,
        path: PurePath,
    ) -> None:
        self._root = root_attribute  # type: ignore
        self._path = path

    @property
    def allowed_extensions(self) -> tp.FrozenSet[str]:
        return self._allowed_extensions

    @classmethod
    @exceptions.handle_unknown_exception
    def load_from_path(
        cls,
        path: tp.Union[str, PurePath] = ".env",
        encoding: str = "utf-8",
    ) -> FileConfigType:
        load_dotenv(dotenv_path=path, encoding=encoding)
        env_config = EnvConfig.load_from_environment()
        return cls(
            root_attribute=env_config._root,
            path=path if isinstance(path, PurePath) else PurePath(path),
        )

    @classmethod
    @exceptions.handle_unknown_exception
    async def load_from_path_async(
        cls,
        path: tp.Union[str, PurePath] = ".env",
        encoding: str = "utf-8",
    ) -> "FileConfigType":
        return cls.load_from_path(
            path=path,
            encoding=encoding,
        )


BASE_FILE_CONFIG_TYPES: tp.Final[tp.Tuple[tp.Type[FileConfigType], ...]] = (
    YamlConfig,
    JsonConfig,
    TomlConfig,
    IniConfig,
    DotenvConfig,
)



import abc
import functools
import typing as tp

from rxconf import observer


class MetaRxConfDI(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def inject_conf(self) -> tp.Callable:
        pass


class MetaAsyncRxConfDI(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    async def inject_conf(self) -> tp.Callable:
        pass


class InjectableRxConf(MetaRxConfDI):

    def __init__(self, conf: observer.MetaObserver) -> None:
        self._conf = conf

    def inject_conf(self) -> tp.Callable[[tp.Callable[..., tp.Any]], tp.Callable[..., tp.Any]]:
        def decorator(function: tp.Callable[..., tp.Any]) -> tp.Callable[..., tp.Any]:
            @functools.wraps(function)
            def wrapper(*args: tp.Any, **kwargs: tp.Any) -> tp.Any:
                return function(*args, **kwargs, conf=self._conf.get_actual_config())

            return wrapper
        return decorator


class InjectableAsyncRxConf(MetaAsyncRxConfDI):
    def __init__(self, conf: observer.MetaAsyncObserver) -> None:
        self._conf = conf

    async def inject_conf(
            self,
    ) -> tp.Callable[
        [tp.Callable[..., tp.Awaitable[tp.Any]]],
        tp.Callable[..., tp.Awaitable[tp.Any]],
    ]:
        def decorator(
            function: tp.Callable[..., tp.Awaitable[tp.Any]]
        ) -> tp.Callable[..., tp.Awaitable[tp.Any]]:
            @functools.wraps(function)
            async def wrapper(*args: tp.Any, **kwargs: tp.Any) -> tp.Any:
                return await function(
                    *args, **kwargs, conf=await self._conf.get_actual_config()
                )

            return wrapper

        return decorator



import abc
import pathlib
import typing as tp

from rxconf import config_resolver, config_types
from rxconf.config_builder import FileConfigBuilder


class MetaTree(metaclass=abc.ABCMeta):

    def __init__(
        self: "MetaTree",
        config: config_types.ConfigType,
    ) -> None:
        self._config = config

    @abc.abstractmethod
    def __getattr__(self, item: str) -> tp.Any:
        pass


class MetaRxConf(MetaTree, metaclass=abc.ABCMeta):

    @classmethod
    @abc.abstractmethod
    def from_file(
        cls: tp.Type["MetaRxConf"],
        config_path: tp.Union[str, pathlib.PurePath],
        encoding: str,
        file_config_resolver: config_resolver.FileConfigResolver,
    ) -> "MetaRxConf":
        pass

    @classmethod
    @abc.abstractmethod
    def from_env(
        cls: tp.Type["MetaRxConf"],
        prefix: tp.Optional[str] = None,
        remove_prefix: tp.Optional[bool] = False,
    ) -> "MetaRxConf":
        pass


class AsyncMetaRxConf(MetaTree, metaclass=abc.ABCMeta):

    @classmethod
    @abc.abstractmethod
    async def from_file(
        cls: tp.Type["AsyncMetaRxConf"],
        config_path: tp.Union[str, pathlib.PurePath],
        encoding: str,
        file_config_resolver: config_resolver.FileConfigResolver,
    ) -> "AsyncMetaRxConf":
        pass

    @classmethod
    @abc.abstractmethod
    async def from_env(
        cls: tp.Type["AsyncMetaRxConf"],
        prefix: tp.Optional[str] = None,
        remove_prefix: tp.Optional[bool] = False,
    ) -> "AsyncMetaRxConf":
        pass


class RxConf(MetaRxConf):

    def __init__(self, config: config_types.ConfigType) -> None:
        super().__init__(config)

    @classmethod
    def from_file(
        cls: tp.Type["RxConf"],
        config_path: tp.Union[str, pathlib.PurePath],
        encoding: str = "utf-8",
        file_config_resolver: config_resolver.FileConfigResolver = config_resolver.DefaultFileConfigResolver,
    ) -> "RxConf":
        return cls(
            config=FileConfigBuilder(
                config_resolver=file_config_resolver,
            ).build(
                path=config_path,
                encoding=encoding,
            )
        )

    @classmethod
    def from_env(
        cls: tp.Type["RxConf"],
        prefix: tp.Optional[str] = None,
        remove_prefix: tp.Optional[bool] = False,
    ) -> "RxConf":
        return cls(
            config=config_types.EnvConfig.load_from_environment(
                prefix=prefix,
                remove_prefix=remove_prefix,
            ),
        )

    def __getattr__(self, item: str) -> tp.Any:
        return getattr(self._config, item.lower())

    def __repr__(self) -> str:
        return repr(self._config)


class AsyncRxConf(AsyncMetaRxConf):

    def __init__(self, config: config_types.ConfigType) -> None:
        super().__init__(config)

    @classmethod
    async def from_file(
        cls: tp.Type["AsyncRxConf"],
        config_path: tp.Union[str, pathlib.PurePath],
        encoding: str = "utf-8",
        file_config_resolver: config_resolver.FileConfigResolver = config_resolver.DefaultFileConfigResolver,
    ) -> "AsyncRxConf":
        return cls(
            config=await FileConfigBuilder(
                config_resolver=file_config_resolver,
            ).build_async(
                path=config_path,
                encoding=encoding,
            )
        )

    @classmethod
    async def from_env(
        cls: tp.Type["AsyncRxConf"],
        prefix: tp.Optional[str] = None,
        remove_prefix: tp.Optional[bool] = False,
    ) -> "AsyncRxConf":
        return cls(
            config=config_types.EnvConfig.load_from_environment(
                prefix=prefix,
                remove_prefix=remove_prefix,
            ),
        )

    def __getattr__(self, item: str) -> tp.Any:
        return getattr(self._config, item.lower())

    def __repr__(self) -> str:
        return repr(self._config)
